/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * This file is part of SableCC.                             *
 * See the file "LICENSE" for copyright information and the  *
 * terms and conditions for copying, distribution and        *
 * modification of SableCC.                                  *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

package org.sablecc.sablecc;

import java.util.*;

final class LR1ItemSet implements Cloneable, Comparable<LR1ItemSet>
{
  private final Map<LR1Item, LR1Item> items;
  private int hashCode = 0;

  LR1ItemSet()
  {
    this.items = new TreeMap<>();
  }

  private LR1ItemSet(LR1ItemSet set
                      )
  {
    this.items = new TreeMap<>(set.items);
    this.hashCode = set.hashCode;
  }

  void set
    (LR1Item item)
  {
    if(items.put(item, item) == null)
    {
      hashCode += item.hashCode();
      modified_ = true;
    }
  }

  boolean get
    (LR1Item item)
  {
    return items.get(item) != null;
  }

  LR1Item[] items_;
  boolean modified_ = true;

  private void computeArray()
  {
    items_ = items.keySet().toArray(new LR1Item[0]);
    modified_ = false;
  }

  LR1Item[] items()
  {
    if(modified_)
    {
      computeArray();
    }

    return items_;
  }

  @Override
  public String toString()
  {
    String nl = System.getProperty("line.separator");

    StringBuilder result = new StringBuilder();
    result.append("{");
    result.append(nl);
    result.append("\t");

    Production[] productions = Production.productions();
    Symbol[] terminals = Symbol.terminals();
    boolean comma = false;
    for(int i = 0; i < productions.length; i++)
    {
      int rightsideLength = productions[i].rightside().length;

      for(int j = 0; j <= rightsideLength; j++)
      {
        LR0Item lr0Item = new LR0Item(productions[i].index, j);

        for(int k = 0; k < terminals.length; k++)
        {
          LR1Item item = new LR1Item(lr0Item, terminals[k].index);
          if(get
              (item))
          {
            if(comma)
            {
              result.append(",");
              result.append(nl);
              result.append("\t");
            }
            else
            {
              comma = true;
            }

            result.append(item);
          }
        }
      }
    }

    result.append(nl);
    result.append("}");
    return result.toString();
  }

  public String toString(Symbol lookahead)
  {
    String nl = System.getProperty("line.separator");

    LR1Item[] items = items();
    int length = items.length;

    Set<String> strings = new TreeSet<>();

    for(int i = 0; i < length; i++)
    {

      String s = items[i].toString(lookahead);
      //System.out.println("s = " + s + " " + " items[i] : " + items[i] + " lookahead : " + lookahead);
      if(s != null)
      {
        strings.add(s);
      }
    }

    StringBuilder result = new StringBuilder();
    result.append("{");

    boolean colon = false;
    for(String s : strings)
    {
      if(colon)
      {
        result.append(",");
        result.append(nl);
      }
      else
      {
        colon = true;
        result.append(nl);
      }

      result.append("\t");
      result.append(s);
    }

    result.append(nl);
    result.append("}");

    return result.toString();
  }

  @Override
  public LR1ItemSet clone()
  {
    return new LR1ItemSet(this);
  }

  @Override
  public boolean equals(Object obj)
  {
    if((obj == null) ||
        (obj.getClass() != this.getClass()))
    {
      return false;
    }

    LR1ItemSet set
      = (LR1ItemSet) obj;

    return items.keySet().equals(set.items.keySet());
  }

  @Override
  public int hashCode()
  {
    return hashCode;
  }

  @Override
  public int compareTo(LR1ItemSet set)
  {
    int result = items.size() - set.items.size();

    if(result == 0)
    {
      Iterator<LR1Item> e = items.keySet().iterator();
      Iterator<LR1Item> f = set.items.keySet().iterator();

      while(e.hasNext() && f.hasNext() && (result == 0))
      {
        result = e.next().compareTo(f.next());
      }

      if(result == 0)
      {
        if(e.hasNext())
        {
          return 1;
        }

        if(f.hasNext())
        {
          return -1;
        }
      }
    }

    return result;
  }
}
